--[[
EnhancedMixerWagons

Specialization for an enhanced mixer wagon's

Author: 	Ifko[nator]
Date: 		23.03.2024
Version:	1.0

History:	v1.0 @23.03.2024 - initial implemation in FS 22
]]

EnhancedMixerWagons = {};

EnhancedMixerWagons.currentModName = "";
EnhancedMixerWagons.currentModDirectory = "";
EnhancedMixerWagons.maizePlusModTable = "";
EnhancedMixerWagons.xmlSchema = {};

for _, mod in pairs(g_modManager.mods) do
	if mod.title:upper() == "ENHANCED MIXER WAGONS" or mod.title:upper() == "VERBESSERTE FUTTERMISCHWAGEN" then		
		if g_modIsLoaded[mod.modName] then	
			EnhancedMixerWagons.currentModName = mod.modName;
			EnhancedMixerWagons.currentModDirectory = mod.modDir;

			break;
		end;
	end;
end;

for _, mod in pairs(g_modManager.mods) do
	if mod.title:upper() == "MAIZEPLUS" then		
		if g_modIsLoaded[mod.modName] then	
			EnhancedMixerWagons.maizePlusModTable = _G[mod.modName];

			break;
		end;
	end;
end;

local function registerForageMixingEffects()
    local xmlFile = XMLFile.load("motionPathXML", Utils.getFilename("data/effects/mixerWagon/motionPathEffects.xml", EnhancedMixerWagons.currentModDirectory), EnhancedMixerWagons.xmlSchema);

    if xmlFile ~= nil then
        g_motionPathEffectManager:loadMotionPathEffects(xmlFile.handle, "motionPathEffects.motionPathEffect", EnhancedMixerWagons.currentModDirectory, EnhancedMixerWagons.currentModName);

        xmlFile:delete();
    end;
end;

MotionPathEffectManager.loadMapData = Utils.appendedFunction(MotionPathEffectManager.loadMapData, registerForageMixingEffects);

function EnhancedMixerWagons.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(MixerWagon, specializations) and SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations);
end;

function EnhancedMixerWagons.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"onUpdate"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, EnhancedMixerWagons);
	end;
end;

function EnhancedMixerWagons.registerFunctions(vehicleType)
	local newFunctions = {
		"calculatePercentage",
        "resetMixingTime",
        "setMixingTime"
	};
	
	for _, newFunction in ipairs(newFunctions) do
		SpecializationUtil.registerFunction(vehicleType, newFunction, EnhancedMixerWagons[newFunction]);
	end;
end;

function EnhancedMixerWagons:onLoad(savegame)
	local specMixerWagon = self.spec_mixerWagon;
	local specFillUnit = self.spec_fillUnit;

	if specMixerWagon.baleTriggers ~= nil then
		for baleTriggerNumber = 1, #specMixerWagon.baleTriggers do
			local baleTrigger = specMixerWagon.baleTriggers[baleTriggerNumber];

			if not baleTrigger.needsSetIsTurnedOn then
				baleTrigger.needsSetIsTurnedOn = true;
			end;
		end;
	end;

	local modDesc = loadXMLFile("modDesc", Utils.getFilename("modDesc.xml", EnhancedMixerWagons.currentModDirectory));

    specMixerWagon.requiredMixingTime = Utils.getNoNil(getXMLInt(modDesc, "modDesc.enhancedMixerWagon#mixingTime"), 120); --## in seconds, 0 = NO MIXING TIME!
	
	delete(modDesc);

	specMixerWagon.timeIconFilename = Utils.getFilename("hud/hud_mixingTime.png", EnhancedMixerWagons.currentModDirectory); 
    
	specMixerWagon.mixedOutputFillType = FillType.FORAGE;
    specMixerWagon.failedOutputFillType = FillType.FORAGE_FAILED;
    
	specMixerWagon.currentMixingTime = 0;
    specMixerWagon.mixingTimePercentage = 0;
    
	specMixerWagon.isFirstRun = true;
	specMixerWagon.needTurnOn = true;

	specMixerWagon.forceUpdate = false;
	specMixerWagon.isForageOk = false;
	specMixerWagon.isSingleFilled = false;
	
	local uiScale = g_gameSettings:getValue("uiScale");
    local width, height = getNormalizedScreenValues(312 * uiScale, 12 * uiScale);
    
    specMixerWagon.statusBarColorGood = MixerWagonHUDExtension.COLOR.STATUS_BAR_GOOD;
    specMixerWagon.statusBarColorAlmostGood = {0.6, 0.5, 0, 1};
    specMixerWagon.statusBarColorBetter = {1, 0.2, 0, 1};
    specMixerWagon.statusBarColorBad = MixerWagonHUDExtension.COLOR.STATUS_BAR_BAD;
    specMixerWagon.statusBar = StatusBar.new(g_baseUIFilename, g_colorBgUVs, nil, MixerWagonHUDExtension.COLOR.STATUS_BAR_BG, specMixerWagon.statusBarColorBad, nil, 0, 0, width, height);

    if specMixerWagon.timeIconFilename ~= "" then
        width, height = getNormalizedScreenValues(30 * uiScale, 30 * uiScale);

        specMixerWagon.timeOverlay = Overlay.new(specMixerWagon.timeIconFilename, 0, 0, width, height);
    end;

	for _, fillType in ipairs({specMixerWagon.mixedOutputFillType, specMixerWagon.failedOutputFillType}) do
		if not g_fillTypeManager:getIsFillTypeInCategory(fillType, "MIXERWAGON") then
			g_fillTypeManager:addFillTypeToCategory(fillType, g_fillTypeManager.nameToCategoryIndex["MIXERWAGON"]);

			specFillUnit.fillUnits[1].supportedFillTypes[fillType] = true;
		end;
	end;

	if savegame ~= nil then
        local currentMixingTime = getXMLInt(savegame.xmlFile.handle, savegame.key .. ".enhancedMixerWagons#currentMixingTime");
 
        specMixerWagon.forceUpdate = currentMixingTime ~= nil;

        if currentMixingTime ~= nil then
			self:setMixingTime(currentMixingTime);
        end;
    end;
end;
	
function EnhancedMixerWagons:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local specMixerWagon = self.spec_mixerWagon;
    local specMotorized = self.spec_motorized;
	local specFillVolume = self.spec_fillVolume;

    local tipState = self:getTipState();
    local isDischarging = tipState == Trailer.TIPSTATE_OPENING or tipState == Trailer.TIPSTATE_OPEN;
	local fillLevel = self:getFillUnitFillLevel(specMixerWagon.fillUnitIndex);
	local fillType = self:getFillUnitFillType(specMixerWagon.fillUnitIndex);

    specMixerWagon.isFirstRun = false;

	if fillLevel <= 0 and specMixerWagon.mixingTimePercentage > 0 then
		--## reset mixing time if the forage mixer is empty

		self:resetMixingTime();
	end;

	if fillType == specMixerWagon.mixedOutputFillType and specMixerWagon.mixingTimePercentage < 1 then
		--## set the mixing time to max, if the forage mixer gets filled with the mixedOutputFillType
		
		self:setMixingTime(specMixerWagon.requiredMixingTime * 1000);
	end;

	if specMixerWagon.forceUpdate or self:getIsTurnedOn() or isDischarging then
		self:raiseActive();
		

		if not isDischarging then
			if fillLevel > 0 and not specMixerWagon.isSingleFilled then
				if specMixerWagon.requiredMixingTime > 0 and specMixerWagon.isForageOk or #specMixerWagon.mixerWagonFillTypes == 1 then
					local maxClamp = 0;

					if not specMixerWagon.isSingleFilled or #specMixerWagon.mixerWagonFillTypes == 1 then
						maxClamp = (specMixerWagon.requiredMixingTime * 1000) / 3;

						if specMixerWagon.currentMixingTime > maxClamp then
							maxClamp = specMixerWagon.currentMixingTime;
						end;
					end;

					if specMixerWagon.isForageOk or #specMixerWagon.mixerWagonFillTypes == 1 then
						maxClamp = specMixerWagon.requiredMixingTime * 1000;
					end;

					local currentMixingTime = MathUtil.clamp((specMixerWagon.currentMixingTime + dt), 0, maxClamp);

					self:setMixingTime(currentMixingTime);
				end;

				local fillType = specMixerWagon.failedOutputFillType;

				if specMixerWagon.mixingTimePercentage == 1 or specMixerWagon.requiredMixingTime == 0 then
					fillType = specMixerWagon.mixedOutputFillType;

					if specMotorized ~= nil then
						specMotorized.motorStopTimerDuration = specMotorized.motorStopTimerDurationBackUp;
					end;
				else
					if specMotorized ~= nil then
						specMotorized.motorStopTimerDuration = math.huge;
					end;
				end;

				if fillType ~= nil then
					self:setFillUnitFillType(specMixerWagon.fillUnitIndex, fillType);
					self:addFillUnitFillLevel(self:getOwnerFarmId(), specMixerWagon.fillUnitIndex, 0, fillType, ToolType.UNDEFINED, nil);

					if specFillVolume.volumes[specMixerWagon.fillUnitIndex] ~= nil then
						--## Hardcore overwrite for the fill volume texture... Otherwhise, it will only update the texture after start discharging ...

						local textureArrayIndex = g_fillTypeManager:getTextureArrayIndexByFillTypeIndex(fillType);

						if textureArrayIndex ~= nil then
							setShaderParameter(specFillVolume.volumes[specMixerWagon.fillUnitIndex].volume, "fillTypeId", textureArrayIndex - 1, 0, 0, 0, false);
						end;
					end;
				end;
			else
				if specMotorized ~= nil then
					specMotorized.motorStopTimerDuration = specMotorized.motorStopTimerDurationBackUp;
				end;
			end;
		end;
	end;

	specMixerWagon.forceUpdate = false;
end;

function EnhancedMixerWagons:calculatePercentage()
    local specMixerWagon = self.spec_mixerWagon;
    
    specMixerWagon.mixingTimePercentage = specMixerWagon.currentMixingTime / (specMixerWagon.requiredMixingTime * 1000);
end;

function EnhancedMixerWagons:resetMixingTime()
    self:setMixingTime(0);
    self:calculatePercentage();
end;

function EnhancedMixerWagons:setMixingTime(currentMixingTime)
    local specMixerWagon = self.spec_mixerWagon;
    
    if currentMixingTime ~= specMixerWagon.currentMixingTime then
       specMixerWagon.currentMixingTime = currentMixingTime;
    end;

	self:calculatePercentage();
end;

function EnhancedMixerWagons:saveToXMLFile(xmlFile, key, usedModNames)
    setXMLInt(xmlFile.handle, key .. "#currentMixingTime", self.spec_mixerWagon.currentMixingTime);
end;

function EnhancedMixerWagons:addFillUnitFillLevel(superFunc, farmId, fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData)
	local specMixerWagon = self.spec_mixerWagon;

	if fillUnitIndex ~= specMixerWagon.fillUnitIndex then
		return superFunc(self, farmId, fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData);
	end;

	local mixerWagonFillType = specMixerWagon.fillTypeToMixerWagonFillType[fillTypeIndex];
	local oldFillLevel = self:getFillUnitFillLevel(fillUnitIndex);

	if EnhancedMixerWagons.maizePlusModTable ~= "" then
		--## do maize plus stuff
		local fillTypeMost = specMixerWagon.fillTypeMost or FillType.UNKNOWN;
    	local fillTypeMostLevel = 0;

    	if fillLevelDelta >= 0 then
			for _, fillType in pairs(specMixerWagon.mixerWagonFillTypes) do     
				if fillType.fillLevel > fillTypeMostLevel then
					for fillTypeIndex, _ in pairs(fillType.fillTypes) do
						fillTypeMost = fillTypeIndex;

						break;
					end;

					fillTypeMostLevel = fillType.fillLevel;
				end;
			end;
		else
			local fillLevel = 0;

			if specMixerWagon.fillTypeToMixerWagonFillType[specMixerWagon.fillTypeMost] ~= nil and specMixerWagon.fillTypeToMixerWagonFillType[specMixerWagon.fillTypeMost].fillLevel ~= nil then
				fillLevel = specMixerWagon.fillTypeToMixerWagonFillType[specMixerWagon.fillTypeMost].fillLevel;
			end;
			
			if fillLevel == 0 then
				fillTypeMost = FillType.UNKNOWN;
				fillTypeMostLevel = 0;
			end;
		end;

		specMixerWagon.fillTypeMost = fillTypeMost;

		if specMixerWagon.outputFillType ~= nil then
			--## set output fill type to the selected output fill type from maize plus

			specMixerWagon.mixedOutputFillType = FillType[specMixerWagon.outputFillType];
		end;
	end;

	if fillTypeIndex == specMixerWagon.mixedOutputFillType then
		--## allow fill the mixedOutputFillType in the forage mixer, if acceppted, without resetting the mixing time.

		if oldFillLevel == 0 then
			fillLevelDelta = math.max(fillLevelDelta, -oldFillLevel);

			local newFillLevel = 0;

			for _, entry in pairs(specMixerWagon.mixerWagonFillTypes) do
				--## split the mixedOutputFillType to the ingredients of the recept for the hud in the F1 menu
				
				local entryPercentage = (entry.maxPercentage * fillLevelDelta + 0.01) / 2;
				local entryDelta = fillLevelDelta * entryPercentage;

				entry.fillLevel = math.max(entry.fillLevel + entryDelta, 0);

				newFillLevel = newFillLevel + entry.fillLevel;
			end;
		end;

		self:raiseDirtyFlags(specMixerWagon.dirtyFlag);

		return superFunc(self, farmId, fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData);
	end;

	if fillTypeIndex == specMixerWagon.mixedOutputFillType and fillLevelDelta > 0 then
		for _, entry in pairs(specMixerWagon.mixerWagonFillTypes) do
			local delta = fillLevelDelta * entry.ratio;

			self:addFillUnitFillLevel(farmId, fillUnitIndex, delta, next(entry.fillTypes), toolType, fillPositionData);
		end;

		return fillLevelDelta;
	end;

	if mixerWagonFillType == nil then
		if fillLevelDelta < 0 and oldFillLevel > 0 then
			fillLevelDelta = math.max(fillLevelDelta, -oldFillLevel);

			local newFillLevel = 0;

			for _, entry in pairs(specMixerWagon.mixerWagonFillTypes) do
				local entryDelta = fillLevelDelta * entry.fillLevel / oldFillLevel;

				entry.fillLevel = math.max(entry.fillLevel + entryDelta, 0);

				newFillLevel = newFillLevel + entry.fillLevel;
			end

			if newFillLevel < 0.1 then
				for _, entry in pairs(specMixerWagon.mixerWagonFillTypes) do
					entry.fillLevel = 0;
				end;

				fillLevelDelta = -oldFillLevel;
			end;

			if newFillLevel <= 0 and specMixerWagon.requiredMixingTime > 0 then
                self:resetMixingTime();
            end;

			self:raiseDirtyFlags(specMixerWagon.dirtyFlag);

			return superFunc(self, farmId, fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData);
		end;
		
		if specMixerWagon.mixingTimePercentage == 1 or specMixerWagon.isFirstRun then
			return 0;
		else
			return superFunc(self, farmId, fillUnitIndex, 0, specMixerWagon.failedOutputFillType, toolType, fillPositionData);
		end;
	end;

	local capacity = self:getFillUnitCapacity(fillUnitIndex);
	local free = capacity - oldFillLevel;

	if fillLevelDelta > 0 then
		mixerWagonFillType.fillLevel = mixerWagonFillType.fillLevel + math.min(free, fillLevelDelta);

		if self:getIsSynchronized() then
			specMixerWagon.activeTimer = specMixerWagon.activeTimerMax;
		end;

		specMixerWagon.forceUpdate = true;

        if specMixerWagon.requiredMixingTime > 0 and fillTypeIndex ~= specMixerWagon.mixedOutputFillType then
            if not specMixerWagon.isFirstRun then
                self:resetMixingTime();
            end;
        end;
	else
		mixerWagonFillType.fillLevel = math.max(0, mixerWagonFillType.fillLevel + fillLevelDelta);
	end;

	local newFillLevel = 0;

	for _, fillType in pairs(specMixerWagon.mixerWagonFillTypes) do
		newFillLevel = newFillLevel + fillType.fillLevel;
	end;

	newFillLevel = MathUtil.clamp(newFillLevel, 0, self:getFillUnitCapacity(fillUnitIndex));

	local newFillType = FillType.UNKNOWN;
	local isSingleFilled = false;
	local isForageOk = false;

	if #specMixerWagon.mixerWagonFillTypes > 1 then
		for _, fillType in pairs(specMixerWagon.mixerWagonFillTypes) do
			if newFillLevel == fillType.fillLevel then
				isSingleFilled = true;

				newFillType = next(mixerWagonFillType.fillTypes);

				break;
			end;
		end;
	else
		if specMixerWagon.requiredMixingTime == 0 then    
            newFillType = specMixerWagon.mixedOutputFillType;
        else
            newFillType = specMixerWagon.failedOutputFillType;
        end;

        isSingleFilled = true;
	end;

	if not isSingleFilled then
		isForageOk = true;

		for _, fillType in pairs(specMixerWagon.mixerWagonFillTypes) do
			if fillType.fillLevel < fillType.minPercentage * newFillLevel - 0.01 or fillType.fillLevel > fillType.maxPercentage * newFillLevel + 0.01 then
				isForageOk = false;

				break;
			end;
		end;
	end;

	specMixerWagon.isForageOk = isForageOk;
	specMixerWagon.isSingleFilled = isSingleFilled;

	if isForageOk then
        if specMixerWagon.requiredMixingTime == 0 then    
            newFillType = specMixerWagon.mixedOutputFillType;
        else
            newFillType = specMixerWagon.failedOutputFillType;
        end;
	elseif not isSingleFilled then
		newFillType = specMixerWagon.failedOutputFillType;
    end;

	self:raiseDirtyFlags(specMixerWagon.dirtyFlag);
	self:setFillUnitFillType(fillUnitIndex, newFillType);
	
	if EnhancedMixerWagons.maizePlusModTable ~= "" then
		--## not sure what the reason for this code is, but maize plus had it, so i simply added it..
		if newFillLevel <= -fillLevelDelta then
			for _, entry in pairs(specMixerWagon.mixerWagonFillTypes) do
				entry.fillLevel = 0;
			end;

			newFillLevel = 0;
		end;
	end;

	return superFunc(self, farmId, fillUnitIndex, newFillLevel - oldFillLevel, newFillType, toolType, fillPositionData);
end;

function EnhancedMixerWagons:getDischargeFillType(superFunc, dischargeNode)
	local specMixerWagon = self.spec_mixerWagon;

    if fillUnitIndex == specMixerWagon.fillUnitIndex then
        local currentFillType = self:getFillUnitFillType(fillUnitIndex);
        local fillLevel = self:getFillUnitFillLevel(fillUnitIndex);
		local fillUnitIndex = dischargeNode.fillUnitIndex;

		if specMixerWagon.mixingTimePercentage < 1 and fillLevel > 0 then
			return specMixerWagon.failedOutputFillType, 1;
		end;
        
        local preselectedFillType = FillType.UNKNOWN;
        local preselectedFillTypeLevel = 0;

        if not specMixerWagon.isForageOk and fillLevel > 0 then 
	    	if specMixerWagon.fillTypeMost ~= nil and specMixerWagon.fillTypeToMixerWagonFillType ~= nil and specMixerWagon.fillTypeToMixerWagonFillType[specMixerWagon.fillTypeMost] ~= nil then 
        		preselectedFillType = next(specMixerWagon.fillTypeToMixerWagonFillType[specMixerWagon.fillTypeMost].fillTypes);
        		preselectedFillTypeLevel = specMixerWagon.fillTypeToMixerWagonFillType[specMixerWagon.fillTypeMost].fillLevel;
    		end;

    		if preselectedFillType ~= FillType.UNKNOWN and preselectedFillTypeLevel > 0 then
    			currentFillType = preselectedFillType;
		  	else
				specMixerWagon.fillTypeMost = FillType.UNKNOWN;

		  		currentFillType = FillType.UNKNOWN;

				for _, entry in pairs(specMixerWagon.mixerWagonFillTypes) do
					if entry.fillLevel > 0 then
						currentFillType = next(entry.fillTypes);

						specMixerWagon.fillTypeMost = currentFillType;
						
						break;
					end;
				end	;
			end;
        end;

        return currentFillType, 1;
    end;

    return superFunc(self, dischargeNode);
end;

MixerWagon.addFillUnitFillLevel = EnhancedMixerWagons.addFillUnitFillLevel;
MixerWagon.getDischargeFillType = EnhancedMixerWagons.getDischargeFillType;

if EnhancedMixerWagons.maizePlusModTable ~= "" then
	--## restore the 'addFillUnitFillLevel' and 'getDischargeFillType' function's back to default, since i have added the changes there for maize plus
	
	EnhancedMixerWagons.maizePlusModTable.mixerWagonRecipeChanger.addFillUnitFillLevel = MixerWagon.addFillUnitFillLevel;
	EnhancedMixerWagons.maizePlusModTable.mixerWagonRecipeChanger.getDischargeFillType = MixerWagon.getDischargeFillType;
end;

function EnhancedMixerWagons:mixerWagonDraw(superFunc, leftPosX, rightPosX, posY)
	setTextColor(unpack(self.uiTextColor));
	setTextBold(true);
	setTextAlignment(RenderText.ALIGN_LEFT);
	renderText(leftPosX, posY + self.displayHeight - self.uiTextSize * 1.5, self.uiTextSize, ("%s (%s)"):format(g_i18n:getText("info_mixingRatio"), self.vehicle:getFullName()));
	setTextBold(false);

	local totalFillLevel = 0;
	local specMixerWagon = self.mixerWagon;

	if self.vehicle:getFillUnitFillLevel(specMixerWagon.fillUnitIndex) > 0 then
		for fillTypeNumber, mixerWagonFillType in ipairs(specMixerWagon.mixerWagonFillTypes) do
			totalFillLevel = totalFillLevel + mixerWagonFillType.fillLevel;

			self.fillTypeStatus[fillTypeNumber].fillLevel = mixerWagonFillType.fillLevel;
		end;
	end;

	local barY = posY + #self.fillTypeStatus * self.helpHeightPerFruit + self.helpHeightPerFruit * 2.6;

	for _, fillTypeDisplay in ipairs(self.fillTypeStatus) do
		local percentage = 0;

		barY = barY - self.helpHeightPerFruit;

		if self.vehicle:getFillUnitFillLevel(specMixerWagon.fillUnitIndex) > 0 then
			percentage = fillTypeDisplay.fillLevel / totalFillLevel;
		end;

		if fillTypeDisplay.fillLevel > 0 then
			if self.vehicle:getFillUnitFillType(specMixerWagon.fillUnitIndex) ~= specMixerWagon.failedOutputFillType or fillTypeDisplay.minPercentage <= percentage and percentage <= fillTypeDisplay.maxPercentage then
				fillTypeDisplay.statusBar:setColor(unpack(fillTypeDisplay.statusBarColor));
			else
				fillTypeDisplay.statusBar:setColor(unpack(fillTypeDisplay.statusBarColor2));
			end;
		end;

		fillTypeDisplay.statusBar:setPosition(rightPosX - fillTypeDisplay.statusBar.width, barY + (self.helpHeightPerFruit - fillTypeDisplay.statusBar.height) * 0.5);
		fillTypeDisplay.statusBar:setValue(percentage);
		fillTypeDisplay.statusBar:render();

		if fillTypeDisplay.overlay ~= nil then
			fillTypeDisplay.overlay:setPosition(leftPosX, barY + (self.helpHeightPerFruit - fillTypeDisplay.overlay.height) * 0.5);
			fillTypeDisplay.overlay:render();
		end;

		setTextAlignment(RenderText.ALIGN_RIGHT);
		renderText(leftPosX + self.fillLevelTextOffsetX, barY + self.fillLevelTextOffsetY, self.uiTextSize, ("%1.1f%%"):format(percentage * 100));
		setTextAlignment(RenderText.ALIGN_LEFT);

		local y = fillTypeDisplay.statusBar.y + fillTypeDisplay.statusBar.height * 0.5;
		local maxLeft = fillTypeDisplay.statusBar.x + fillTypeDisplay.statusBar.width;
		local xLeft = math.min(fillTypeDisplay.statusBar.x + fillTypeDisplay.statusBar.width * fillTypeDisplay.minPercentage, maxLeft);
		local xRight = math.min(fillTypeDisplay.statusBar.x + fillTypeDisplay.statusBar.width * fillTypeDisplay.maxPercentage, maxLeft);

		self.fillRangeMarkerOverlay:setInvertX(false);
		self.fillRangeMarkerOverlay:setPosition(xLeft, y);
		self.fillRangeMarkerOverlay:render();
		self.fillRangeMarkerOverlay:setInvertX(true);
		self.fillRangeMarkerOverlay:setPosition(xRight - self.fillRangeMarkerOverlay.width, y);
		self.fillRangeMarkerOverlay:render();
	end;

	barY = posY + self.helpHeightPerFruit * 2;

	if Utils.getNoNil(specMixerWagon.requiredMixingTime, 0) > 0 then
        local function getFormattedTime()
            local function getExtraZero(time, skip)    
                if time < 10 and not skip then
                    time = "0" .. time;
                end;
				
                return tostring(time);
            end;
			
			local neededTime = MathUtil.clamp(math.ceil(specMixerWagon.requiredMixingTime - (specMixerWagon.currentMixingTime / 1000)), 0, specMixerWagon.requiredMixingTime);

            local hours = getExtraZero(math.floor(neededTime / 3600), true);
            local minutes = getExtraZero(math.floor((neededTime % 3600) / 60), hours == "0");
            local seconds = getExtraZero(math.floor(neededTime % 60), minutes == "0");

			local text_hours = g_i18n:getText("text_hours", EnhancedMixerWagons.currentModName);
			local text_minutes = g_i18n:getText("text_minutes", EnhancedMixerWagons.currentModName);
			local text_seconds = g_i18n:getText("text_seconds", EnhancedMixerWagons.currentModName);

			if hours == "1" then
				text_hours = text_hours:sub(1, text_hours:len() - 1);
			end;

			if minutes == "01" or minutes == "1" then
				text_minutes = text_minutes:sub(1, text_minutes:len() - 1);
			end;

			if seconds == "01" or seconds == "1" then
				text_seconds = text_seconds:sub(1, text_seconds:len() - 1);
			end;
        
            local formatedTime = hours .. " " .. text_hours .. " : " .. minutes .. " " .. text_minutes .. " : " .. seconds .. " " .. text_seconds;

			if hours == "0" then
				formatedTime = minutes .. " " .. text_minutes .. " : " .. seconds .. " " .. text_seconds;

				if minutes == "0" then
					formatedTime = seconds .. " " .. text_seconds;
				end;
			end;

			return formatedTime;
        end;
		
		setTextBold(true);
        renderText(leftPosX, barY - (self.helpHeightPerFruit - (self.uiTextSize * 1.35)), self.uiTextSize, g_i18n:getText("NEEDED_MIXING_TIME") .. " " .. getFormattedTime());
        setTextBold(false);
		
        if specMixerWagon.timeOverlay ~= nil then
            specMixerWagon.timeOverlay:setPosition(leftPosX, barY - (self.helpHeightPerFruit - self.uiTextSize) * 3.8);
            specMixerWagon.timeOverlay:render();
        end;

		local mixingTimePercentage = specMixerWagon.mixingTimePercentage;

		setTextAlignment(RenderText.ALIGN_RIGHT);
		renderText(leftPosX + self.fillLevelTextOffsetX, barY - (self.helpHeightPerFruit - self.uiTextSize) * 3.3, self.uiTextSize, ("%1.1f%%"):format(mixingTimePercentage * 100));
		setTextAlignment(RenderText.ALIGN_LEFT);

        if mixingTimePercentage == 1 then
            specMixerWagon.statusBar:setColor(unpack(specMixerWagon.statusBarColorGood));
        elseif mixingTimePercentage > 0.7 and mixingTimePercentage < 1 then
            specMixerWagon.statusBar:setColor(unpack(specMixerWagon.statusBarColorAlmostGood));
        elseif mixingTimePercentage > 0.35 and mixingTimePercentage < 0.7 then
            specMixerWagon.statusBar:setColor(unpack(specMixerWagon.statusBarColorBetter));
        else
            specMixerWagon.statusBar:setColor(unpack(specMixerWagon.statusBarColorBad));
        end;

        specMixerWagon.statusBar:setPosition(rightPosX - specMixerWagon.statusBar.width, barY - (self.helpHeightPerFruit - specMixerWagon.statusBar.height) * 2.4);
        specMixerWagon.statusBar:setValue(mixingTimePercentage);
        specMixerWagon.statusBar:render();

		local y = specMixerWagon.statusBar.y + specMixerWagon.statusBar.height * 0.5;
		local maxLeft = specMixerWagon.statusBar.x + specMixerWagon.statusBar.width;
		local xLeft = math.min(specMixerWagon.statusBar.x + specMixerWagon.statusBar.width * 0, maxLeft);
		local xRight = math.min(specMixerWagon.statusBar.x + specMixerWagon.statusBar.width * math.max(mixingTimePercentage, 0.05), maxLeft);

		self.fillRangeMarkerOverlay:setInvertX(false);
		self.fillRangeMarkerOverlay:setPosition(xLeft, y);
		self.fillRangeMarkerOverlay:render();
		self.fillRangeMarkerOverlay:setInvertX(true);
		self.fillRangeMarkerOverlay:setPosition(xRight - self.fillRangeMarkerOverlay.width, y);
		self.fillRangeMarkerOverlay:render();

        self.displayHeight = (#specMixerWagon.mixerWagonFillTypes + 3.5) * self.helpHeightPerFruit + self.helpHeightOffset; --## extend help window for mixing time satus bar
    else
        self.displayHeight = (#specMixerWagon.mixerWagonFillTypes + 1) * self.helpHeightPerFruit + self.helpHeightOffset; --## restore help window size for mixing wagons without time satus bar
    end;

	return posY;
end;

MixerWagonHUDExtension.draw = Utils.overwrittenFunction(MixerWagonHUDExtension.draw, EnhancedMixerWagons.mixerWagonDraw);

function EnhancedMixerWagons:getFillTypesFromXML(superFunc, fillTypes, usedFillTypes, xmlFile, key)
	local fillTypeNameStr = xmlFile:getValue(key);

	if fillTypeNameStr == nil then
		Logging.xmlError(xmlFile, "Missing fillTypes for ingredient '%s'", key);

		return false;
	end;

	if fillTypeNameStr:upper():find("POTATO_CUT") and EnhancedMixerWagons.maizePlusModTable ~= "" then
		--## add failed forage to food group moisture food (maize plus)

		fillTypeNameStr = fillTypeNameStr .. " FORAGE_FAILED";
	elseif fillTypeNameStr:upper():find("GRASS_WINDROW") and not fillTypeNameStr:upper():find("DRY") and not fillTypeNameStr:upper():find("WET") then
		--## add failed forage to food group grass
		
		fillTypeNameStr = fillTypeNameStr .. " FORAGE_FAILED";
	end;

	local fillTypeNames = fillTypeNameStr:split(" ");

	for _, fillTypeName in pairs(fillTypeNames) do
		local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(fillTypeName);

		if fillTypeIndex ~= nil then
			if usedFillTypes[fillTypeIndex] == nil then
				table.addElement(fillTypes, fillTypeIndex);
			else
				Logging.xmlWarning(xmlFile, "FillType '%s' already used in other ingredient", fillTypeName);
			end;
		else
			Logging.xmlWarning(xmlFile, "FillType '%s' not defined. Ignoring it", fillTypeName);
		end;
	end;

	if #fillTypes == 0 then
		Logging.xmlError(xmlFile, "No fillTypes defined - '%s'", key);

		return false;
	end;

	return true;
end;

AnimalFoodSystem.getFillTypesFromXML = Utils.overwrittenFunction(AnimalFoodSystem.getFillTypesFromXML, EnhancedMixerWagons.getFillTypesFromXML);